from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit


class IT1500(ControlUnit):
    _repetitions = 6
    _timebase = 275
    _pausedata = 5600  # not really needed, just for keeping reenginering data

    from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
    from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

    def __init__(self, manufacturer: Manufacturer = Manufacturer.INTERTECHNO,
                 model: ControlUnitModel = ControlUnitModel.IT_1500):
        super().__init__(manufacturer, model)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            'CODE': '[01]{26}$',
            'UNIT': '^([1-9]|0[0-9]|1[1-6])$'
        }

    def get_pulse_data(self, action: Action):
        _sho = 1
        _lon = 5
        _d0 = [(_sho, _sho), (_sho, _lon)]
        _d1 = [(_sho, _lon), (_sho, _sho)]
        _d = (_d0, _d1)  # for bit builders
        _pre = (_sho, 10)
        _post = [(_sho, 41)]

        tuples = [_pre]

        cfg = self.get_channel_config()
        for bit in cfg['CODE']:
            tuples += _d1 if bit == '1' else _d0

        tuples += _d0  # all

        if action is Action.ON:
            tuples += _d1
        elif action is Action.OFF:
            tuples += _d0

        unit = int(cfg['UNIT'])
        for i in range(4):
            tuples += _d1 if unit & 1 << i else _d0

        tuples += _post

        return tuples, self._repetitions, self._timebase



from raspyrfm_client.device_implementations.controlunit.manufacturer.intertechno.IT1500 import IT1500


class Hama121938(IT1500):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super().__init__(Manufacturer.HAMA, ControlUnitModel.MODEL_00121938)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit
from raspyrfm_client.device_implementations.gateway.manufacturer.gateway_constants import GatewayModel
from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer


class Gateway(object):
    """
    Base gateway implementation
    """

    def __init__(self, manufacturer: Manufacturer, model: GatewayModel, host: str, port: int):
        self._manufacturer = manufacturer
        self._model = model
        self._firmware_version = None

        if host is None:
            host = "127.0.0.1"
        self._host = host

        if port is None:
            port = 49880
        self._port = port

    def get_manufacturer(self) -> Manufacturer:
        """
        :return: the gateway manufacturer
        """
        return self._manufacturer

    def get_model(self) -> GatewayModel:
        """
        :return: the gateway model
        """
        return self._model

    def get_firmware_version(self) -> str:
        """
        :return: the gateway firmware version
        """
        return self._firmware_version

    def get_host(self) -> str:
        """
        :return: the ip/host address of the gateway (if one was found or specified manually)
        """
        return self._host

    def get_port(self) -> int:
        """
        :return: port of the gateway 
        """
        return self._port

    def get_search_response_regex_literal(self) -> str:
        """
        :return: a regular expression that matches the response to a "search" broadcast 
        """
        raise NotImplementedError

    @staticmethod
    def create_from_broadcast(host: str, message: str):
        """
        :param host: the host that sent the message
        :param message: the search response message
        :return: a new instance of this gateway based on a broadcast response message
        """
        raise NotImplementedError

    def generate_code(self, device: ControlUnit, action: Action) -> str:
        raise NotImplementedError



from raspyrfm_client.device_implementations.controlunit.manufacturer.noname.RSL366 import RSL366


class FSL100(RSL366):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super(RSL366, self).__init__(Manufacturer.M_E,
                                     ControlUnitModel.FLS100)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit


class Ec000x(ControlUnit):
    _on = ['0', '1']
    _off = ['1', '0']

    _lo = 1
    _hi = 3
    _seqLo = [(_lo, _hi)]
    _seqHi = [(_hi, _lo)]
    _sync = [(1, 31)]

    _timebase = 3300
    _pauselen = 5600

    _chvalues = [
        _seqHi + _seqHi + _seqHi,
        _seqLo + _seqHi + _seqHi,
        _seqHi + _seqLo + _seqHi,
        _seqHi + _seqHi + _seqLo
    ]

    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super().__init__(Manufacturer.LOGILINK, ControlUnitModel.EC000X)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF, Action.PAIR]

    def get_channel_config_args(self):
        return {
            'CODE': '^[0-9A-F]{5}$',
            'CH': '^[1-4]$'
        }

    def get_pulse_data(self, action: Action):
        cfg = self.get_channel_config()
        tuples = []
        for nibble in cfg['CODE']:
            val = int(nibble, 16)
            for i in range(4):
                if (val & 0x8) > 0:
                    tuples += self._seqHi
                else:
                    tuples += self._seqLo
                val <<= 1

        if action is Action.ON:
            tuples += self._seqHi
            repetitions = 5
        elif action is Action.PAIR:
            tuples += self._seqHi
            repetitions = 15
        elif action is Action.OFF:
            tuples += self._seqLo
            repetitions = 5

        tuples += self._chvalues[int(cfg['CH']) - 1]

        tuples += self._sync

        return tuples, repetitions, self._timebase



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class ZtcS316A(HX2262Compatible):
    _h = '0'
    _l = 'f'
    _on = ['0', '1']
    _off = ['1', '0']
    _repetitions = 5

    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super().__init__(Manufacturer.WESTFALIA, ControlUnitModel.ZTC_S316A)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            'A': '^[01]$',
            'B': '^[01]$',
            'C': '^[01]$',
            'D': '^[01]$',
            'E': '^[01]$',
            'F': '^[01]$',
            'CH': '^[1-4]$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        for i in range(6):
            bits += self._h if cfg[chr(i + ord('A'))] == '1' else self._l

        for i in range(4):
            bits += self._h if int(cfg['CH']) == (4 - i) else self._l

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off

        return bits, self._repetitions



"""
Base class for all controlunit implementations
"""

import re

from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer


class ControlUnit(object):
    def __init__(self, manufacturer: Manufacturer, model: ControlUnitModel):
        self._manufacturer = manufacturer
        self._model = model
        self._channel = None

    def __str__(self):
        return ("Manufacturer: " + self._manufacturer.value + "\n" +
                "Model: " + self._model.value + "\n" +
                "Supported Actions: " + str(self.get_supported_actions()) + "\n" +
                "Channel: " + str(self.get_channel_config()))

    def get_manufacturer(self) -> Manufacturer:
        """
        :return: the device manufacturer
        """
        return self._manufacturer

    def get_model(self) -> ControlUnitModel:
        """
        :return: the device model
        """
        return self._model

    def set_channel_config(self, **channel_arguments) -> None:
        """
        Sets the channel as multiple arguments.
        See implementation specific details about how the channel should be passed in.

        :param channel_arguments:
        """
        argchecks = self.get_channel_config_args()
        for arg in argchecks:
            if arg not in channel_arguments:
                raise ValueError("arguments should contain key \"" + arg + "\"")
            if re.match(argchecks[arg], str(channel_arguments[arg])) is None:
                raise ValueError("argument \"" + arg + "\" out of range, does not match to " + argchecks[arg])

        self._channel = channel_arguments

    def get_channel_config_args(self):
        """
        gets required config arguments and their regular expression to check the erguments
        has to be implemented by inheriting classes
        
        :return: dictionary of arguments
        
        example: {"ID": "^[A-F]$", "CH": "^[1-4]$"}
        """
        raise NotImplementedError

    def get_channel_config(self) -> dict or None:
        """
        :return: the channel setup as a dict
        """
        return self._channel

    def get_supported_actions(self) -> [Action]:
        """
        :return: the supported actions of this device
        """
        raise NotImplementedError

    def get_pulse_data(self, action: Action):
        """
        generates pulse data
        :return: (pulse pairs, repetitions, timebase)
        """
        raise NotImplementedError



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class Rcs14G(HX2262Compatible):
    _on = ['f', 'f']
    _off = ['f', '0']

    _codes = [
        ['0', '0', '0', 'f', 'f', '0', 'f', '0', 'f', 'f'],
        ['0', '0', '0', 'f', 'f', 'f', '0', '0', 'f', 'f'],
        ['0', '0', '0', 'f', '0', 'f', 'f', '0', 'f', 'f'],
        ['0', '0', '0', '0', 'f', 'f', 'f', '0', 'f', 'f']
    ]

    _repetitions = 5

    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super().__init__(Manufacturer.LUX_GMBH, ControlUnitModel.RCS_14G)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            'CH': '^[1-4]$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []
        bits += self._codes[int(cfg['CH']) - 1]

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off
        else:
            raise ValueError("Invalid action")

        return bits, self._repetitions



import unittest
from builtins import print

import rstr

from raspyrfm_client import RaspyRFMClient
from raspyrfm_client.device_implementations.gateway.manufacturer.gateway_constants import GatewayModel
from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer


class TestStringMethods(unittest.TestCase):
    def test_random_controlunit_config(self):
        """
        Tests all device_implementations with random configurations.
        """

        rfm_client = RaspyRFMClient()

        from raspyrfm_client.device_implementations.controlunit.base import ControlUnit

        def test_device(device: ControlUnit):
            """
            Tests random device_implementations configurations for the specified device_implementations
            
            :param device: the device_implementations to test 
            """

            self.assertIsNotNone(device.get_manufacturer())
            self.assertIsNotNone(device.get_model())
            self.assertIsNotNone(device.get_supported_actions())

            channel_config_args = device.get_channel_config_args()

            # tests 50 randomly chosen configurations
            for i in range(50):
                channel_config = {}

                for arg in channel_config_args:
                    channel_config[arg] = rstr.xeger(channel_config_args[arg])

                device.set_channel_config(**channel_config)

                # create gateway instance
                gateway = rfm_client.get_gateway(Manufacturer.INTERTECHNO, GatewayModel.ITGW)

                for action in device.get_supported_actions():
                    generated_code = gateway.generate_code(device, action)
                    self.assertIsNotNone(generated_code)

        def test_models(manufacturer: Manufacturer):
            """
            Tests all models of the specified manufacturer
            
            :param manufacturer:  manufacturer to test all available models
            """
            for model in rfm_client.get_supported_controlunit_models(manufacturer):
                print("Testing " + model.value)
                device = rfm_client.get_controlunit(manufacturer, model)
                test_device(device)

        for manufacturer in rfm_client.get_supported_controlunit_manufacturers():
            test_models(manufacturer)

        print("All random config device_implementations tests passed!")

    def test_gateway_init(self):

        rfm_client = RaspyRFMClient()

        from raspyrfm_client.device_implementations.gateway.base import Gateway

        def test_gateway(gateway: Gateway):
            self.assertIsNotNone(gateway)

        def test_models(manufacturer: Manufacturer):
            for model in rfm_client.get_supported_gateway_models(manufacturer):
                gateway = rfm_client.get_gateway(manufacturer, model)
                test_gateway(gateway)

        for manufacturer in rfm_client.get_supported_gateway_manufacturers():
            test_models(manufacturer)


if __name__ == '__main__':
    unittest.main()



from raspyrfm_client.device_implementations.controlunit.manufacturer.bat.RC3500_A_IP44_DE import RC3500_A_IP44_DE


class RC_AAA1000_A_IP44_Outdoor(RC3500_A_IP44_DE):
    def __init__(self):
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        super().__init__(Manufacturer.BAT, ControlUnitModel.RC_AAA1000_A_IP44_Outdoor)



from raspyrfm_client.device_implementations.controlunit.manufacturer.brennenstuhl.RCS1000NComfort import RCS1000NComfort


class RCS1044NComfort(RCS1000NComfort):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super(RCS1000NComfort, self).__init__(Manufacturer.BRENNENSTUHL,
                                              ControlUnitModel.RCS_1044_N_COMFORT)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit
from raspyrfm_client.device_implementations.gateway.base import Gateway


class ConnAir(Gateway):

    def __init__(self, host: str = None, port: int = 49880):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.gateway.manufacturer.gateway_constants import GatewayModel
        super().__init__(Manufacturer.SIMPLE_SOLUTIONS, GatewayModel.CONNAIR, host, port)

    @staticmethod
    def create_from_broadcast(host: str, message: str):
        # manufacturer = message[message.index('VC:') + 3:message.index(';MC')]
        # model = message[message.index('MC:') + 3:message.index(';FW')]
        firmware_version = message[message.index('FW:') + 3:message.index(';IP')]
        # parsed_host = message[message.index('IP:') + 3:message.index(';;')]

        instance = ConnAir(host)
        instance._firmware_version = firmware_version

        return instance

    def get_search_response_regex_literal(self) -> str:
        return "HCGW:.*VC:Simple Solutions;MC:.*;FW:.+;IP:.+;;"

    def generate_code(self, device: ControlUnit, action: Action) -> str:
        """
        This method can be implemented by inheriting classes if it does not implement get_pulse_data
        :param device: The device to generate the code for
        :param action: action to execute
        :return: signal code
        """
        if device.get_channel_config() is None:
            raise ValueError("Missing channel configuration :(")
        if action not in device.get_supported_actions():
            raise ValueError("Unsupported action: " + str(action))

        pulsedata = device.get_pulse_data(action)
        _head_connair = "TXP:0,0,"
        _code = _head_connair
        _code += str(pulsedata[1]) + ','  # add repetitions
        _code += str(5600) + ','
        _code += str(pulsedata[2]) + ','  # add timebase

        _code = _code + str(len(pulsedata[0])) + ','
        for pulse in pulsedata[0]:
            _code += str(pulse[0]) + ','
            _code += str(pulse[1]) + ','
        return _code[:-1]



"""
Example usage of the RaspyRFMClient can be found in the example.py file
"""
import socket

from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit
from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
from raspyrfm_client.device_implementations.gateway.base import Gateway
from raspyrfm_client.device_implementations.gateway.manufacturer.gateway_constants import GatewayModel
from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer


class RaspyRFMClient:
    """
    This class is the main interface for generating and sending signals.
    """

    """
    This dictionary maps manufacturer and model constants to their implementation class.
    It is filled automatically when creating an instance of this class.
    """
    _GATEWAY_IMPLEMENTATIONS_DICT = {}
    _CONTROLUNIT_IMPLEMENTATIONS_DICT = {}

    def __init__(self):
        """
        Creates a new client object.
        """
        self.reload_implementation_classes()

    def reload_implementation_classes(self):
        """
        Dynamically reloads device implementations 
        """
        print("Loading implementation classes...")
        self._reload_gateway_implementations()
        self._reload_controlunit_implementations()
        print("Done!")

    @staticmethod
    def __import_submodules(package, recursive=True):
        """ Import all submodules of a module, recursively, including subpackages

        :param package: package (name or actual module)
        :param recursive: loads all subpackages
        :type package: str | module
        :rtype: dict[str, types.ModuleType]
        """
        import pkgutil
        import importlib

        if isinstance(package, str):
            package = importlib.import_module(package)
        results = {}
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
            full_name = package.__name__ + '.' + name
            results[full_name] = importlib.import_module(full_name)
            if recursive and is_pkg:
                results.update(RaspyRFMClient.__import_submodules(full_name))
        return results

    @staticmethod
    def __get_all_subclasses(base_class):
        """
        Returns a list of all currently imported classes that are subclasses (even multiple levels)
        of the specified base class.
        :param base_class: base class to match classes to
        :return: list of classes
        """
        all_subclasses = []

        for subclass in base_class.__subclasses__():
            all_subclasses.append(subclass)
            all_subclasses.extend(RaspyRFMClient.__get_all_subclasses(subclass))

        return all_subclasses

    def _reload_gateway_implementations(self) -> None:
        """
        Finds gateway implementations in the "raspyrfm_client.device_implementations.gateway.manufacturer" package.
        This works by recursively searching through supbackages and finding classes that have
        the gateway base class (base.py) as a superclass.
        """
        self._GATEWAY_IMPLEMENTATIONS_DICT = {}

        from raspyrfm_client.device_implementations.gateway import manufacturer
        RaspyRFMClient.__import_submodules(manufacturer)

        for gateway_implementation in RaspyRFMClient.__get_all_subclasses(Gateway):
            # ignore classes that are disabled by the developer
            if hasattr(gateway_implementation, "DISABLED") and gateway_implementation.DISABLED is True:
                continue

            gateway_instance = gateway_implementation()
            brand = gateway_instance.get_manufacturer()
            model = gateway_instance.get_model()

            if brand not in self._GATEWAY_IMPLEMENTATIONS_DICT:
                self._GATEWAY_IMPLEMENTATIONS_DICT[brand] = {}

            self._GATEWAY_IMPLEMENTATIONS_DICT[brand][model] = gateway_implementation

    def _reload_controlunit_implementations(self) -> None:
        """
        Finds device implementations in the "raspyrfm_client.device_implementations.manufacturer" package.
        This works by recursively searching through supbackages and finding classes that have
        the device base class (base.py) as a superclass.
        """

        self._CONTROLUNIT_IMPLEMENTATIONS_DICT = {}

        from raspyrfm_client.device_implementations.controlunit import manufacturer
        RaspyRFMClient.__import_submodules(manufacturer)

        for device_implementation in RaspyRFMClient.__get_all_subclasses(ControlUnit):
            # ignore classes that are disabled by the developer
            if hasattr(device_implementation, "DISABLED") and device_implementation.DISABLED is True:
                continue

            device_instance = device_implementation()
            brand = device_instance.get_manufacturer()
            model = device_instance.get_model()

            if brand not in self._CONTROLUNIT_IMPLEMENTATIONS_DICT:
                self._CONTROLUNIT_IMPLEMENTATIONS_DICT[brand] = {}

            self._CONTROLUNIT_IMPLEMENTATIONS_DICT[brand][model] = device_implementation

    def get_supported_gateway_manufacturers(self):
        """
        :return: a list of supported gateway manufacturers
        """
        return self._GATEWAY_IMPLEMENTATIONS_DICT.keys()

    def get_supported_gateway_models(self, manufacturer: Manufacturer) -> [GatewayModel]:
        """
        :param manufacturer: supported gateway manufacturer
        :return: a list of supported gateway models for this gateway manufacturer
        """
        return self._GATEWAY_IMPLEMENTATIONS_DICT[manufacturer].keys()

    def get_gateway(self, manufacturer: Manufacturer, model: GatewayModel, host: str = None,
                    port: int = None) -> Gateway:
        """
        Use this method to get a gateway implementation instance
        :param manufacturer: gateway manufacturer
        :param model: gateway model
        :param host: gateway host address (optional)
        :param port: gateway port (optional)
        :return: gateway implementation
        """
        return self._GATEWAY_IMPLEMENTATIONS_DICT[manufacturer][model](host, port)

    def get_supported_controlunit_manufacturers(self) -> [str]:
        """
        :return: a list of supported control unit manufacturers
        """
        return self._CONTROLUNIT_IMPLEMENTATIONS_DICT.keys()

    def get_supported_controlunit_models(self, manufacturer: Manufacturer) -> [ControlUnitModel]:
        """
        :param manufacturer: supported control unit manufacturer
        :return: a list of supported control unit models for this manufacturer
        """
        return self._CONTROLUNIT_IMPLEMENTATIONS_DICT[manufacturer].keys()

    def get_controlunit(self, manufacturer: Manufacturer, model: ControlUnitModel) -> ControlUnit:
        """
        Use this method to get a device implementation intance
        :param manufacturer: device manufacturer
        :param model: device model
        :return: device implementation
        """
        return self._CONTROLUNIT_IMPLEMENTATIONS_DICT[manufacturer][model]()

    def list_supported_gateways(self) -> None:
        """
        Prints an indented list of all supported manufacturers and models
        """
        for manufacturer in self._GATEWAY_IMPLEMENTATIONS_DICT:
            print(manufacturer.value)
            for model in self._GATEWAY_IMPLEMENTATIONS_DICT[manufacturer].keys():
                print("  " + model.value)

    def list_supported_controlunits(self) -> None:
        """
        Prints an indented list of all supported manufacturers and models
        """
        for manufacturer in self._CONTROLUNIT_IMPLEMENTATIONS_DICT:
            print(manufacturer.value)
            for model in self._CONTROLUNIT_IMPLEMENTATIONS_DICT[manufacturer].keys():
                print("  " + model.value)

    def search(self) -> [Gateway]:
        """
        Sends a local network broadcast with a specified message.
        If a gateway is present it will respond to this broadcast.

        If a valid response is found the properties of this client object will be updated accordingly.

        :return: list of gateways
        """

        import re
        import socket
        from socket import AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_REUSEADDR, SO_BROADCAST

        found_gateways = []
        all_gateways = []

        # get all gateway implementations in a list
        for manufacturer in self.get_supported_gateway_manufacturers():
            for model in self.get_supported_gateway_models(manufacturer):
                all_gateways.append(self.get_gateway(manufacturer, model))

        # send the broadcast
        cs = socket.socket(AF_INET, SOCK_DGRAM)
        cs.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        cs.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)

        _broadcast_message = b'SEARCH HCGW'

        cs.sendto(_broadcast_message, ('255.255.255.255', 49880))

        cs.setblocking(True)
        cs.settimeout(1)

        # receive the message(s)
        try:
            while True:
                data, address = cs.recvfrom(4096)
                # print("Received message: \"%s\"" % data)
                # print("Address: " + address[0])

                message = data.decode()

                # for each device implementation, check if the response matches the expected pattern
                # and add an instance of this gateway implementation to the found_gateways list
                for gateway in all_gateways:
                    if re.match(gateway.get_search_response_regex_literal(), message) is not None:
                        found_gateways.append(gateway.create_from_broadcast(address[0], message))

        except socket.timeout:
            return found_gateways
        finally:
            return found_gateways

    def send(self, gateway: Gateway, device: ControlUnit, action: Action) -> None:
        """
        Use this method to generate codes for actions on supported device.
        It will generates a string that can be interpreted by the the RaspyRFM module.
        The string contains information about the rc signal that should be sent.

        :param gateway: the gateway to generate the code for
        :param device: the device to generate the code for
        :param action: action to execute
        """

        if gateway.get_host() is None:
            print("Missing host, nothing sent.")
            return

        message = gateway.generate_code(device, action)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP
        sock.sendto(bytes(message, "utf-8"), (gateway.get_host(), gateway.get_port()))



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class RSL366(HX2262Compatible):
    _h = '0'
    _l = 'f'
    _repetitions = 5

    from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
    from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

    def __init__(self, manufacturer: Manufacturer = Manufacturer.NONAME,
                 model: ControlUnitModel = ControlUnitModel.RSL366):
        super().__init__(manufacturer, model)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            'CODE': '^[1-4]$',
            'CH': '^[1-4]$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        bits += self.calc_match_bits(int(cfg['CODE']) - 1, 4, (self._l, self._h))
        bits += self.calc_match_bits(int(cfg['CH']) - 1, 4, (self._l, self._h))

        bits += [self._l, self._l, self._l]  # fixed

        if action is Action.ON:
            bits += [self._l]
        elif action is Action.OFF:
            bits += [self._h]
        else:
            raise ValueError("Invalid action")

        return bits, self._repetitions



from raspyrfm_client.device_implementations.controlunit.manufacturer.elro.AB440S import AB440S


class FSS31000W(AB440S):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
        super(AB440S, self).__init__(Manufacturer.VIVANCO, ControlUnitModel.FSS31000W)



from raspyrfm_client.device_implementations.controlunit.manufacturer.elro.AB440S import AB440S


class AB440ID(AB440S):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super(AB440S, self).__init__(Manufacturer.ELRO, ControlUnitModel.AB440ID)

    def get_channel_config_args(self):
        return {
            '1': '^[01]$',
            '2': '^[01]$',
            '3': '^[01]$',
            '4': '^[01]$',
            '5': '^[01]$',
            'CH': '^[A-C]$'  # device has only switches A-C
        }



from raspyrfm_client.device_implementations.controlunit.manufacturer.intertechno.CMR1000 import CMR1000


class ITR300(CMR1000):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super(CMR1000, self).__init__(Manufacturer.INTERTECHNO, ControlUnitModel.ITR_300)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel
from raspyrfm_client.device_implementations.controlunit.manufacturer.brennenstuhl.RCS1000NComfort import RCS1000NComfort
from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer


class RC3500_A_IP44_DE(RCS1000NComfort):
    _l = 'f'
    _h = '0'
    _on = [_h, _l]
    _off = [_l, _h]
    _repetitions = 5

    def __init__(self, manufacturer: Manufacturer = Manufacturer.BAT,
                 model: ControlUnitModel = ControlUnitModel.RC3500_A_IP44_DE):
        super(RCS1000NComfort, self).__init__(manufacturer, model)

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        for i in range(5):
            bits += self._h if cfg[str(i + 1)] == '1' else self._l

        ch = ord(cfg['CH']) - ord('A')
        bits += self.calc_match_bits(ch, 5, (self._l, self._h))

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off

        return bits, self._repetitions



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class Ritter(HX2262Compatible):
    _l = 'f'
    _h = '0'
    _on = [_l, _l]
    _off = [_h, _h]
    _repetitions = 5

    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super(Ritter, self).__init__(Manufacturer.REV, ControlUnitModel.RITTER)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            '1': '^[01]$',
            '2': '^[01]$',
            '3': '^[01]$',
            '4': '^[01]$',
            '5': '^[01]$',
            '6': '^[01]$',
            'CH': '^[A-D]$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        for i in range(6):
            bits.append(self._h if cfg[str(i + 1)] == '1' else self._l)

        for i in range(4):
            bits.append(self._h if cfg['CH'] == chr(i + ord('A')) else self._l)

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off
        else:
            raise ValueError("Invalid action")

        return bits, self._repetitions



from enum import Enum


class Manufacturer(Enum):
    """
    Enum holding all manufacturer names
    """

    BAT = "BAT"

    BRENNENSTUHL = "Brennenstuhl"

    ELRO = "Elro"

    HAMA = "hama"

    INTERTECHNO = "Intertechno"

    INTERTEK = "Intertek"

    LOGILINK = "LogiLink"

    LUX_GMBH = "Emil Lux GmbH"

    MUMBI = "Mumbi"

    M_E = "m-e GmbH modern-electronics"

    NONAME = "noname"

    POLLIN_ELECTRONIC = "Pollin Electronic"

    REV = "REV"

    SEEGEL_SYSTEME = "Seegel Systeme"

    SIMPLE_SOLUTIONS = "Simple Solutions"

    UNIVERSAL = "Universal"

    VIVANCO = "Vivanco"

    WESTFALIA = "Westfalia"



from raspyrfm_client.device_implementations.controlunit.manufacturer.intertechno.IT1500 import IT1500


class Telecontrol2(IT1500):
    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super(IT1500, self).__init__(Manufacturer.REV, ControlUnitModel.TELECONTROL8342LC)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class CMR1000(HX2262Compatible):
    _l = '0'
    _h = 'f'
    _on = [_h, _h]
    _off = [_h, _l]
    _repetitions = 5

    def __init__(self):
        from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
        from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

        super().__init__(Manufacturer.INTERTECHNO, ControlUnitModel.CMR_1000)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            'master': '^[A-P]$',
            'slave': '^([1-9]|[0][1-9]|[1][1-6])$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        master = ord(cfg['master']) - ord('A')
        bits += self.calc_int_bits(master, 4, (self._l, self._h))

        slave = int(cfg['slave']) - 1
        bits += self.calc_int_bits(slave, 4, (self._l, self._h))

        bits += [self._l, self._h]  # fixed

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off

        return bits, self._repetitions



from setuptools import setup, find_packages

setup(
    name='raspyrfm_client',
    version='1.2.1',
    description='A library to send rc signals with the RaspyRFM module',
    license='GPLv3+',
    author='Markus Ressel',
    author_email='mail@markusressel.de',
    url='https://www.markusressel.de',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.4',
    ],
    install_requires=[
    ],
    tests_require=[
        'rstr'
    ]
)



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.manufacturer.universal.HX2262Compatible import HX2262Compatible


class RCS1000NComfort(HX2262Compatible):
    _l = 'f'
    _h = '0'
    _on = [_l, _l]
    _off = [_l, _h]
    _repetitions = 5

    from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
    from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

    def __init__(self, manufacturer: Manufacturer = Manufacturer.BRENNENSTUHL,
                 model: ControlUnitModel = ControlUnitModel.RCS_1000_N_COMFORT):
        super(HX2262Compatible, self).__init__(manufacturer, model)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON, Action.OFF]

    def get_channel_config_args(self):
        return {
            '1': '^[01]$',
            '2': '^[01]$',
            '3': '^[01]$',
            '4': '^[01]$',
            '5': '^[01]$',
            'CH': '^[A-E]$'
        }

    def get_bit_data(self, action: Action):
        cfg = self.get_channel_config()
        bits = []

        for i in range(5):
            bits += self._h if cfg[str(i + 1)] == '1' else self._l

        ch = ord(cfg['CH']) - ord('A')
        bits += self.calc_match_bits(ch, 5, (self._l, self._h))

        if action is Action.ON:
            bits += self._on
        elif action is Action.OFF:
            bits += self._off

        return bits, self._repetitions



from raspyrfm_client.device_implementations.controlunit.actions import Action
from raspyrfm_client.device_implementations.controlunit.base import ControlUnit


class HX2262Compatible(ControlUnit):
    _sho = 1
    _lon = 3

    _d0 = [(_sho, _lon), (_sho, _lon)]
    _d1 = [(_lon, _sho), (_lon, _sho)]
    _df = [(_sho, _lon), (_lon, _sho)]

    _sync = (_sho, 31)

    _repetitions = 5
    _timebase = 350
    _pausedata = 5600  # not really needed, just for keeping reenginering data

    from raspyrfm_client.device_implementations.manufacturer_constants import Manufacturer
    from raspyrfm_client.device_implementations.controlunit.controlunit_constants import ControlUnitModel

    def __init__(self, manufacturer: Manufacturer = Manufacturer.UNIVERSAL,
                 model: ControlUnitModel = ControlUnitModel.HX2262):
        super().__init__(manufacturer, model)

    def get_supported_actions(self) -> [Action]:
        return [Action.ON]

    def get_channel_config_args(self):
        return {
            '1': '^[01fF]$',
            '2': '^[01fF]$',
            '3': '^[01fF]$',
            '4': '^[01fF]$',
            '5': '^[01fF]$',
            '6': '^[01fF]$',
            '7': '^[01fF]$',
            '8': '^[01fF]$',
            '9': '^[01fF]$',
            '10': '^[01fF]$',
            '11': '^[01fF]$',
            '12': '^[01fF]$'
        }

    def get_pulse_data(self, action: Action):
        bitdata = self.get_bit_data(action)

        # print("bits:", bitdata[0])

        if len(bitdata[0]) != 12:
            raise ValueError("Bits not configured")

        tuples = []

        for bit in bitdata[0]:
            bit_value = bit.lower()
            if bit_value == 'f':
                tuples += self._df
            elif bit_value == '0':
                tuples += self._d0
            elif bit_value == '1':
                tuples += self._d1
            else:
                raise ValueError(
                    "Invalid bit value \"" + bit_value + "\"! Must be one of ['0', '1', 'f'] (case insensitive)")

        tuples.append(self._sync)  # sync pulse

        return tuples, bitdata[1], self._timebase

    def get_bit_data(self, action: Action):
        """
        This method should be implemented by inheriting classes
        :return: char array (12 bits '0'|'1'|'f'), number of repetitions
        """

        cfg = self.get_channel_config()
        bits = []

        for i in range(12):
            bits += cfg[str(i + 1)]

        return bits, self._repetitions

    def calc_int_bits(self, value, num_bits, bitvalues):
        """
        converts an integervalue to a bitarray using binary format
        :param value: integer to convert
        :param num_bits: number of bits to convert
        :param bitvalues: tuples of low- and highvalue to use for the array
        
        :return:
        array if given bitvalues
        """
        bits = []
        for i in range(num_bits):
            bits.append(bitvalues[1] if (value & 1 << i) != 0 else bitvalues[0])
        return bits

    def calc_match_bits(self, value, num_bits, bitvalues):
        """
        converts an integervalue to a bitarray, only bit nr. value will be set
        :param value: integer to convert
        :param num_bits: number of bits to convert
        :param bitvalues: tuples of low- and highvalue to use for the array
        
        :return:
        array if given bitvalues
        """
        bits = []
        for i in range(num_bits):
            bits.append(bitvalues[1] if value == i else bitvalues[0])
        return bits
